##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Implementation of templates.
"""

from zope import component
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.pagetemplate.interfaces import IPageTemplate
from zope.pagetemplate.pagetemplate import PageTemplate

from z3c.template import interfaces


class Macro:
    # XXX: We can't use Zope's `TALInterpreter` class directly
    # because it (obviously) only supports the Zope page template
    # implementation. As a workaround or trick we use a wrapper
    # template.

    wrapper = PageTemplate()
    wrapper.write(
        '<metal:main use-macro="python: options[\'macro\']" />'
    )

    def __init__(self, template, name, view, request, contentType):
        self.macro = template.macros[name]
        self.contentType = contentType
        self.view = view
        self.request = request

    def __call__(self, **kwargs):
        kwargs['macro'] = self.macro
        kwargs.setdefault('view', self.view)
        kwargs.setdefault('request', self.request)
        result = self.wrapper(**kwargs)

        if not self.request.response.getHeader("Content-Type"):
            self.request.response.setHeader(
                "Content-Type", self.contentType)

        return result


class TemplateFactory:
    """Template factory."""

    template = None

    def __init__(self, filename, contentType, macro=None):
        self.contentType = contentType
        self.template = ViewPageTemplateFile(
            filename, content_type=contentType)
        self.macro = macro

    def __call__(self, view, request, context=None):
        if self.macro is None:
            return self.template
        return Macro(
            self.template, self.macro, view, request, self.contentType)


class BoundViewTemplate:
    __self__ = None
    __func__ = None

    def __init__(self, pt, ob):
        object.__setattr__(self, '__func__', pt)
        object.__setattr__(self, '__self__', ob)

    im_self = property(lambda self: self.__self__)
    im_func = property(lambda self: self.__func__)

    def __call__(self, *args, **kw):
        if self.__self__ is None:
            im_self, args = args[0], args[1:]
        else:
            im_self = self.__self__
        return self.__func__(im_self, *args, **kw)

    def __setattr__(self, name, v):
        raise AttributeError("Can't set attribute", name)

    def __repr__(self):
        return "<BoundViewTemplate of %r>" % self.__self__


class ViewTemplate:
    def __init__(self, provides=IPageTemplate, name=''):
        self.provides = provides
        self.name = name

    def __call__(self, instance, *args, **keywords):
        template = component.queryMultiAdapter(
            (instance, instance.request, instance.context),
            self.provides, name=self.name)
        if template is None:
            template = component.getMultiAdapter(
                (instance, instance.request),
                self.provides, name=self.name)
        return template(instance, *args, **keywords)

    def __get__(self, instance, type):
        return BoundViewTemplate(self, instance)


getViewTemplate = ViewTemplate


class GetPageTemplate(ViewTemplate):

    def __init__(self, name=''):
        self.provides = interfaces.IContentTemplate
        self.name = name


getPageTemplate = GetPageTemplate


class GetLayoutTemplate(ViewTemplate):

    def __init__(self, name=''):
        self.provides = interfaces.ILayoutTemplate
        self.name = name


getLayoutTemplate = GetLayoutTemplate



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""ZCML Directives
"""
import os

import zope.component.zcml
import zope.configuration.fields
import zope.interface
import zope.schema
from zope.configuration.exceptions import ConfigurationError
from zope.publisher.interfaces.browser import IDefaultBrowserLayer

import z3c.template.interfaces
from z3c.template.template import TemplateFactory


class ITemplateDirective(zope.interface.Interface):
    """Parameters for the template directive."""

    template = zope.configuration.fields.Path(
        title='Layout template.',
        description="Refers to a file containing a page template (should "
        "end in extension ``.pt`` or ``.html``).",
        required=True,
    )

    name = zope.schema.TextLine(
        title="The name of the template.",
        description="The name is used to look up the template.",
        default='',
        required=False)

    macro = zope.schema.TextLine(
        title='Macro',
        description="""
            The macro to be used.
            This allows us to define different macros in one template.
            The template designer can now create a whole site, the
            ViewTemplate can then extract the macros for single viewlets
            or views.
            If no macro is given the whole template is used for rendering.
            """,
        default='',
        required=False,
    )

    for_ = zope.configuration.fields.GlobalObject(
        title='View',
        description='The view for which the template should be available',
        default=zope.interface.Interface,
        required=False,
    )

    layer = zope.configuration.fields.GlobalObject(
        title='Layer',
        description='The layer for which the template should be available',
        required=False,
        default=IDefaultBrowserLayer,
    )

    context = zope.configuration.fields.GlobalObject(
        title='Context',
        description='The context for which the template should be available',
        required=False,
    )

    provides = zope.configuration.fields.GlobalInterface(
        title="Interface the template provides",
        description="This attribute specifies the interface the template"
        " instance will provide.",
        default=z3c.template.interfaces.IContentTemplate,
        required=False,
    )

    contentType = zope.schema.ASCIILine(
        title='Content Type',
        description='The content type identifies the type of data.',
        default='text/html',
        required=False,
    )


class ILayoutTemplateDirective(ITemplateDirective):
    """Parameters for the layout template directive."""

    provides = zope.configuration.fields.GlobalInterface(
        title="Interface the template provides",
        description="This attribute specifies the interface the template"
        " instance will provide.",
        default=z3c.template.interfaces.ILayoutTemplate,
        required=False,
    )


def templateDirective(
        _context, template, name='',
        for_=zope.interface.Interface, layer=IDefaultBrowserLayer,
        provides=z3c.template.interfaces.IContentTemplate,
        contentType='text/html', macro=None, context=None):

    # Make sure that the template exists
    template = os.path.abspath(str(_context.path(template)))
    if not os.path.isfile(template):
        raise ConfigurationError("No such file", template)

    factory = TemplateFactory(template, contentType, macro)
    zope.interface.directlyProvides(factory, provides)

    if context is not None:
        for_ = (for_, layer, context)
    else:
        for_ = (for_, layer)

    # register the template
    if name:
        zope.component.zcml.adapter(_context, (factory,), provides,
                                    for_, name=name)
    else:
        zope.component.zcml.adapter(_context, (factory,), provides,
                                    for_)


def layoutTemplateDirective(
        _context, template, name='',
        for_=zope.interface.Interface, layer=IDefaultBrowserLayer,
        provides=z3c.template.interfaces.ILayoutTemplate,
        contentType='text/html', macro=None, context=None):

    templateDirective(_context, template, name, for_, layer, provides,
                      contentType, macro, context)



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tests
"""
import doctest
import itertools
import unittest

import z3c.pt
import z3c.ptcompat
from zope.component import testing
from zope.configuration import xmlconfig

import z3c.template.template


def setUp(test):
    testing.setUp(test)
    # Traversal Setup
    from zope.traversing.testing import setUp
    setUp()
    test.globs['root'] = object()


def tearDown(test):
    testing.tearDown(test)


def setUpZPT(suite):
    setUp(suite)


def setUpZ3CPT(suite):
    setUp(suite)
    xmlconfig.XMLConfig('configure.zcml', z3c.pt)()
    xmlconfig.XMLConfig('configure.zcml', z3c.ptcompat)()

    # We have to cook this template explicitly, because it's a module
    # global.
    z3c.template.template.Macro.wrapper._cook()


class TestMacro(unittest.TestCase):

    def test_call_sets_content_type(self):

        class Response:
            def __init__(self):
                self.headers = {}
                self.getHeader = self.headers.get

            def setHeader(self, k, v):
                self.headers[k] = v

        class Request:
            def __init__(self):
                self.response = Response()

        class Template:
            def __init__(self):
                self.macros = {}

        template = Template()
        template.macros['name'] = None
        request = Request()
        macro = z3c.template.template.Macro(template, 'name', None,
                                            request, 'text/html')
        macro.wrapper = lambda **kwargs: None

        macro()
        self.assertEqual(
            'text/html', request.response.getHeader("Content-Type"))


class TestBoundViewTemplate(unittest.TestCase):

    def test_call_no__self__uses_first_arg(self):
        def im_func(*args):
            return args
        bound = z3c.template.template.BoundViewTemplate(im_func, None)

        im_self = 1
        args = (2, 3)
        result = bound(im_self, *args)
        self.assertEqual(result, (1, 2, 3))

    def test_cant_setattr(self):
        bound = z3c.template.template.BoundViewTemplate(None, None)
        with self.assertRaisesRegex(AttributeError, "Can't set attribute"):
            setattr(bound, 'im_func', 42)

    def test_repr(self):
        bound = z3c.template.template.BoundViewTemplate(None, 'im_self')
        self.assertEqual("<BoundViewTemplate of 'im_self'>",
                         repr(bound))

    def test_attributes(self):
        func = object()
        bound = z3c.template.template.BoundViewTemplate(func, self)
        self.assertIs(self, bound.im_self)
        self.assertIs(self, bound.__self__)
        self.assertIs(func, bound.im_func)
        self.assertIs(func, bound.__func__)


def test_suite():
    setups = (setUpZPT, setUpZ3CPT)
    doctests = ((
        doctest.DocFileSuite(
            'README.rst',
            setUp=setUp, tearDown=tearDown,
            optionflags=(doctest.NORMALIZE_WHITESPACE
                         | doctest.ELLIPSIS
                         | doctest.IGNORE_EXCEPTION_DETAIL),
        ),
        doctest.DocFileSuite(
            'zcml.rst',
            setUp=setUp, tearDown=tearDown,
            optionflags=(doctest.NORMALIZE_WHITESPACE
                         | doctest.ELLIPSIS
                         | doctest.IGNORE_EXCEPTION_DETAIL),
        ),
    ) for setUp in setups)
    doctests = list(itertools.chain(*doctests))

    suite = unittest.defaultTestLoader.loadTestsFromName(__name__)
    suite.addTests(doctests)

    return suite
